import examplecurves
from arithmeticmeancurve import FrozenStdExtrapolation, VisualIterationTester

sample_curves = examplecurves.Static.create(family_name="verticallinear0")

VisualIterationTester.plot_extrapolation_test(
    curves=sample_curves, extrapolates=FrozenStdExtrapolation()
)



import examplecurves
from arithmeticmeancurve import FrozenStdExtrapolation, VisualIterationTester

sample_curves = examplecurves.Static.create(family_name="horizontallinear3")

from arithmeticmeancurve import ArithmeticMeanCurve

a_mean_curves = ArithmeticMeanCurve(sample_curves)
# VisualIterationTester.plot_extrapolation_test(
#     curves=a_mean_curve,
#     extrapolates=FrozenStdExtrapolation(
#         use_previous_iteration=True, target_threshold=0.00001
#     ),
# )
# VisualIterationTester.plot_extrapolation_test(
#     curves=a_mean_curve,
#     extrapolates=FrozenStdExtrapolation(
#         use_previous_iteration=True, target_threshold=0.0001
#     ),
# )



import examplecurves
from arithmeticmeancurve import meld_set_of_curves_to_family
from arithmeticmeancurve import _calculate_mean_and_std_curve
from arithmeticmeancurve import ArithmeticMeanCurve


def test_mean_curve_calculation_without_end_cap():
    """
    >>> from doctestprinter import print_pandas
    >>> sample_mean, sample_std = test_mean_curve_calculation_without_end_cap()
    >>> print_pandas(sample_mean, formats="{:>.1f}#{:>.3f}")
    x
    0.0  0.000
    0.2  2.224
    0.2  2.386
    0.2  2.389
    0.3  2.457
    0.3  2.467
    0.5  4.448
    0.5  4.772
    0.5  4.778
    0.5  4.915
    0.5  4.935
    0.7  6.672
    0.7  7.158
    0.7  7.167
    0.8  7.372
    0.8  7.402
    0.9  8.896

    """
    sample_curves = examplecurves.Static.create("verticallinear0")
    sample_family = meld_set_of_curves_to_family(sample_curves)
    sample_family.interpolate(method="index", limit_area="inside", inplace=True)
    sample_family.dropna(inplace=True)

    a_mean_curve = ArithmeticMeanCurve(sample_family)
    mean_curve, mean_std = _calculate_mean_and_std_curve(a_mean_curve=a_mean_curve)
    return mean_curve, mean_std


"""
Issue with VerticalLinear0:
    The mean curve was not cut properly around a value of the arithmetic target point.

Cause:
    Y target value (y end mean) was outside the end cap. It is located in between the
    middle value block section and the end cap.

Solution(s):
    - The is_meet code of the break condition class was hardened.
    - The last middle block value line is concatenated to the extrapolated end cap to
      catch target values in between both.

>>> from arithmeticmeancurve import ArithmeticMeanCurve
>>> import examplecurves
>>> sample_curves = examplecurves.Static.create("VerticalLinear0")
>>> a_mean_curve = ArithmeticMeanCurve(sample_curves)
>>> a_mean_curve.calculate()
"""
import copy

import examplecurves
import pandas

from arithmeticmeancurve import (
    ArithmeticMeanCurve,
    _prepare_end_cap_for_extrapolation,
    _extrapolate_until_targets_are_included,
    _calculate_cutting_index_for_value,
)

_current_test_curves = examplecurves.Static.create("VerticalLinear0")


def test_find_cutting_index():
    """
    >>> import examplecurves
    >>> from doctestprinter import print_pandas
    >>> sample_curves = examplecurves.Static.create("VerticalLinear0")
    >>> end_mean_curve, target_end_y_value = test_find_cutting_index()
    >>> print_pandas(end_mean_curve, formats="{:>.3f}")
    x
    0.907  8.896
    0.972  9.534
    >>> round(target_end_y_value, 6)
    9.532235
    >>> index_for_target_y = _calculate_cutting_index_for_value(
    ...     series_to_cut=end_mean_curve,
    ...     cutting_y_value=target_end_y_value,
    ... )
    >>> round(index_for_target_y, 6)
    0.972307
    """
    sample_curves = copy.copy(_current_test_curves)
    a_mean_curve = ArithmeticMeanCurve(sample_curves)
    target_end_x_value = a_mean_curve.statistics.end_x_mean
    target_end_y_value = a_mean_curve.statistics.end_y_mean

    a_mean_curve.prepare_extrapolation()
    prepared_end_cap = _prepare_end_cap_for_extrapolation(a_mean_curve=a_mean_curve)

    last_full_value_row = a_mean_curve.get_last_full_row()
    last_value_row_mean_value = last_full_value_row.mean(axis=1)
    starting_x_mean_value = last_value_row_mean_value.iloc[0]

    extrapolated_end_cap = _extrapolate_until_targets_are_included(
        extrapolates=a_mean_curve,
        end_cap=prepared_end_cap,
        target_x_mean_value=target_end_x_value,
        target_y_mean_value=target_end_y_value,
        starting_last_mean_value=starting_x_mean_value,
    )
    extended_extrapolated_end_cap = pandas.concat(
        [last_full_value_row, extrapolated_end_cap])
    current_end_mean_curve = extended_extrapolated_end_cap.mean(axis=1)
    return current_end_mean_curve, target_end_y_value


from arithmeticmeancurve import MeanCurvePlotter, ArithmeticMeanCurve
import examplecurves

def test_MeanCurvePlotter_test_plot():
    sample_curves = examplecurves.Static.create("verticallinear0")
    MeanCurvePlotter._test_plot(sample_curves=sample_curves)
    MeanCurvePlotter._test_plot(sample_curves=sample_curves, lower_title="lower title", upper_title="upper title")


def test_MeanCurvePlotter_plot():
    sample_curves = examplecurves.Static.create("verticallinear0")
    sample_mean_curve = ArithmeticMeanCurve(sample_curves)
    MeanCurvePlotter._plot(a_mean_curve=sample_mean_curve)


import examplecurves
from arithmeticmeancurve import ArithmeticMeanCurve

sample_curves = examplecurves.Static.create(
   family_name="nonlinear0",
   predefined_offset=1,
)
a_mean_curve = ArithmeticMeanCurve(sample_curves)
mean_curve = a_mean_curve.mean_curve

import matplotlib.pyplot as plt

fig = plt.figure(figsize=(8, 5), dpi=96)
gspec = fig.add_gridspec(1, 3)
axis = fig.add_subplot(gspec[:, :2])
for index, curve in enumerate(sample_curves):
   current_label = "Sample curve {}".format(index)
   axis.plot(curve, marker="o", markersize=3, label=current_label)

axis.plot(
   a_mean_curve.mean_curve,
   "--ko",
   markersize=5,
   label="Arithmetic mean curve"
)

axis.legend(bbox_to_anchor=(1.05, 1.02), loc='upper left')
plt.show()


import numpy
import examplecurves
from arithmeticmeancurve import extract_end_points, ArithmeticMeanCurve

sample_curves = examplecurves.Static.create(
    family_name="nonlinear0", predefined_offset=1,
)
extrapolating_mean_curve = ArithmeticMeanCurve(curves=sample_curves)
end_points = extract_end_points(extrapolating_mean_curve.family_of_curves)
x_values, y_values = end_points[:, 0], end_points[:, 1]

import matplotlib.pyplot as plt

# Setup figure
fig = plt.figure(figsize=(9, 5), dpi=96)
gspec = fig.add_gridspec(1, 1)
axis = fig.add_subplot(gspec[0, 0])
axis.set_title("extrapolated arithmetic mean curve")

axis.plot(
    *extrapolating_mean_curve.std_bars.horizontal,
    "-|",
    markersize=20,
    label="Horizontal standard deviation",
)
axis.plot(
    *extrapolating_mean_curve.std_bars.vertical,
    "-_",
    markersize=20,
    label="Vertical standard deviation",
)
axis.plot(
    x_values,
    y_values,
    "ok",
    markersize=8,
    label="end points of family of curves",
)
axis.plot(extrapolating_mean_curve.std_circle, "--k", label="Standard deviation circle")
axis.plot(extrapolating_mean_curve.family_of_curves)
# Finishing touch
axis.set_xlim(0.9, 1.4)
axis.set_ylim(8.5, 12.0)

# Finishing touch
axis.legend()
plt.show()


"""
>>> import examplecurves
>>> from arithmeticmeancurve import ArithmeticMeanCurve, FrozenStdExtrapolation
>>> from doctestprinter import print_pandas
>>> a_mean_curve = examplecurves.Static.create("verticallinear0")
>>> sample_mean_curve = ArithmeticMeanCurve(a_mean_curve)
>>> print_pandas(sample_mean_curve.family_of_curves, formats="{:>.4}")
          y_0     y_1    y_2    y_3    y_4
x
0.0000  0.000   0.000  0.000  0.000  0.000
0.2267  2.292   2.256  2.281  2.215  2.076
0.2432  2.459   2.421  2.447  2.377  2.227
0.2435  2.462   2.424  2.450  2.380  2.230
0.2505  2.532   2.493  2.520  2.448  2.293
0.2516  2.543   2.503  2.531  2.458  2.303
0.4535  4.583   4.513  4.562  4.430  4.151
0.4865  4.917   4.841  4.894  4.753  4.454
0.4871  4.924   4.848  4.900  4.759  4.459
0.5010  5.065   4.986  5.041  4.895  4.587
0.5031  5.085   5.007  5.061  4.916  4.606
0.6802  6.875   6.769  6.843  6.646  6.227
0.7297  7.376   7.262  7.341  7.130  6.680
0.7306  7.385   7.271  7.350  7.139  6.689
0.7516  7.597   7.479  7.561  7.343  6.880
0.7547  7.628   7.510  7.592  7.373  6.909
0.9069  9.167   9.025  9.124  8.861  8.302
0.9730    nan   9.683  9.788  9.506  8.907
0.9742    nan   9.695  9.801    nan  8.918
1.0021    nan   9.973    nan    nan  9.174
1.0062    nan  10.014    nan    nan    nan
>>> sample_mean_curve.statistics
FamilyOfCurveStatistics
         end_x     end_y  min_x  min_y     max_x     max_y  start_x  start_y
mean  0.972476  9.532235    0.0    0.0  0.972476  9.532235      0.0      0.0
std   0.039746  0.376375    0.0    0.0  0.039746  0.376375      0.0      0.0
>>> extrapolates = FrozenStdExtrapolation()
>>> extrapolates.prepare_extrapolation(a_mean_curve=sample_mean_curve)
>>> extrapolates
FrozenStdExtrapolation
    Target standard deviation (std): 0.352041900152178
    Curve's relative std positions:
        [ 0.77005112  0.36806549  0.64719569 -0.09924049 -1.68607181]
<BLANKLINE>
"""
import examplecurves

from arithmeticmeancurve import (
    SetOfCurves,
    FrozenStdExtrapolation,
    convert_set_to_family_of_curves,
    FamilyOfCurveStatistics,
    _TargetValuesAreReachedCondition,
    ArithmeticMeanCurve,
)
from arithmeticmeancurve import ExtrapolationIterationTester as EIT
import copy


_current_curves = examplecurves.Static.create("verticallinear0")


def test_iterations():
    """
    >>> from doctestprinter import prepare_pandas, EditingItem, doctest_iter_print
    >>> import examplecurves
    >>> import pandas
    >>> from arithmeticmeancurve import FamilyOfCurveStatistics, convert_set_to_family_of_curves
    >>> iteration_records = test_iterations()
    >>> pandas.options.display.float_format = '{:.4f}'.format
    >>> edits_pandas = EditingItem(prepare_pandas, formats="{:<}{:>.2f}#{:>.4f}")
    >>> doctest_iter_print(iteration_records.to_frames(), edits_item=edits_pandas)
                           y_0     y_1     y_2     y_3     y_4
    sections
    start         0.97     nan  9.6779  9.7833  9.5015  8.9025
    extrapolated  0.97  9.8042     nan     nan     nan     nan
    iteration     0.97  9.1668  9.6779  9.7833  9.5015  8.9025
                  0.97  9.6775  9.6779  9.7833  9.5015  8.9025
                  0.97  9.7796  9.6779  9.7833  9.5015  8.9025
                  0.97  9.8001  9.6779  9.7833  9.5015  8.9025
                  0.97  9.8042  9.6779  9.7833  9.5015  8.9025

    """
    sample_curves = copy.copy(_current_curves)
    frozen_std_extrapolation = FrozenStdExtrapolation()
    iteration_records = EIT.record_iterations(
        set_of_curves=sample_curves, extrapolates=frozen_std_extrapolation
    )
    return iteration_records


def test_break_condition():
    """
    >>> import examplecurves
    >>> test_break_condition()
    [True]
    """
    sample_curves = copy.copy(_current_curves)
    sample_family = convert_set_to_family_of_curves(sample_curves)
    stats = FamilyOfCurveStatistics(family_of_curves=sample_family)

    a_mean_curve = ArithmeticMeanCurve(sample_curves)
    starting_end_mean_value = a_mean_curve.get_last_full_row().mean(axis=1).iloc[0]

    condition = _TargetValuesAreReachedCondition(
        stats.end_x_mean, stats.end_y_mean, starting_end_mean_value
    )

    frozen_std_extrapolation = FrozenStdExtrapolation()
    iteration_records = EIT.record_iterations(
        set_of_curves=sample_curves, extrapolates=frozen_std_extrapolation
    )
    extrapolation_results = iteration_records.get_extrapolation_result()
    condition_results = []
    for x_value, extrapolated_values in extrapolation_results.iterrows():
        condition_is_meet = condition.is_meet(x_value, extrapolated_values)
        condition_results.append(condition_is_meet)
    return condition_results


"""
Issue with VerticalLinear1:
    The mean curve was not cut properly around a value of the arithmetic target point.

Cause:
    The second value of the extrapolated end mean curve is nearest to the targetvalue
    when searched with *_find_positions_in_between*. But the target value is not
    in between its neighbors. The first half of *_find_positions_in_between*
    lead to wrong positions. Therefore the second half of this method.

Solution:
    Edited *_find_positions_in_between*

>>> from doctestprinter import print_pandas
>>> from arithmeticmeancurve import ArithmeticMeanCurve
>>> import examplecurves
>>> a_mean_curve = examplecurves.Static.create("VerticalLinear1")
>>> a_mean_curve = ArithmeticMeanCurve(a_mean_curve)

Test a running calculation.

>>> a_mean_curve.calculate()

Show the results.
>>> print_pandas(a_mean_curve.statistics.stats, formats="{:>}#{:>.4f}")
       end_x    end_y   min_x   min_y   max_x    max_y  start_x  start_y
mean  1.0150  10.1471  0.0000  0.0000  1.0150  10.1471   0.0000   0.0000
 std  0.0125   0.2252  0.0000  0.0000  0.0125   0.2252   0.0000   0.0000
>>> print_pandas(a_mean_curve.mean_curve, formats="{:>.4f}")
x
0.0000   0.0000
0.2495   2.4950
0.2536   2.5355
0.2536   2.5359
0.2536   2.5360
0.2584   2.5832
0.4991   4.9900
0.5072   5.0710
0.5073   5.0718
0.5073   5.0720
0.5167   5.1663
0.7486   7.4850
0.7608   7.6065
0.7609   7.6078
0.7609   7.6079
0.7751   7.7495
0.9981   9.9800
1.0145  10.1418
1.0145  10.1415
1.0150  10.1419
1.0155  10.1471

"""
import copy

import examplecurves
import pandas

from arithmeticmeancurve import (
    ArithmeticMeanCurve,
    _prepare_end_cap_for_extrapolation,
    _extrapolate_until_targets_are_included,
    _calculate_cutting_index_for_value,
    FrozenStdExtrapolation,
    _calculate_mean_and_std_curve,
)


_current_test_curves = examplecurves.Static.create("verticallinear1")


def test_find_cutting_index():
    """
    Debug incorrect cutting index calculation.

    >>> from arithmeticmeancurve import _find_positions_in_between
    >>> import examplecurves
    >>> from doctestprinter import print_pandas
    >>> end_mean_curve, target_end_y_value = test_find_cutting_index()
    >>> round(target_end_y_value, 6)
    10.147148
    >>> print_pandas(end_mean_curve, formats="{:>.4f}")
    x
    0.9981   9.9800
    1.0144  10.1420
    1.0145  10.1418
    1.0145  10.1415
    1.0150  10.1419
    1.0334  10.3160
    >>> _find_positions_in_between(
    ...     source_series=end_mean_curve, search_value=target_end_y_value,
    ... )
    (4, 5)
    >>> index_for_target_y = _calculate_cutting_index_for_value(
    ...     series_to_cut=end_mean_curve,
    ...     cutting_y_value=target_end_y_value,
    ... )
    >>> round(index_for_target_y, 6)
    1.015547
    """
    sample_curves = copy.copy(_current_test_curves)
    a_mean_curve = ArithmeticMeanCurve(sample_curves)
    target_end_x_value = a_mean_curve.statistics.end_x_mean
    target_end_y_value = a_mean_curve.statistics.end_y_mean

    a_mean_curve.prepare_extrapolation()
    prepared_end_cap = _prepare_end_cap_for_extrapolation(a_mean_curve=a_mean_curve)

    last_full_value_row = a_mean_curve.get_last_full_row()
    last_value_row_mean_value = last_full_value_row.mean(axis=1)
    starting_x_mean_value = last_value_row_mean_value.iloc[0]

    extrapolated_end_cap = _extrapolate_until_targets_are_included(
        extrapolates=a_mean_curve,
        end_cap=prepared_end_cap,
        target_x_mean_value=target_end_x_value,
        target_y_mean_value=target_end_y_value,
        starting_last_mean_value=starting_x_mean_value,
    )
    extended_extrapolated_end_cap = pandas.concat(
        [last_full_value_row, extrapolated_end_cap]
    )
    current_end_mean_curve = extended_extrapolated_end_cap.mean(axis=1)
    return current_end_mean_curve, target_end_y_value


def test_find_cutting_index_2():
    """
    Debug incorrect cutting index calculation if FrozenStdExtrapolation is
    run with use_previous_iteration=False.

    >>> from arithmeticmeancurve import _find_positions_in_between
    >>> import examplecurves
    >>> from doctestprinter import print_pandas
    >>> end_mean_curve, target_end_y_value = test_find_cutting_index_2()
    >>> round(target_end_y_value, 6)
    10.147148
    >>> print_pandas(end_mean_curve, formats="{:>.4f}")
    x
    0.9981   9.9800
    1.0144  10.1423
    1.0145  10.1414
    1.0145  10.1401
    1.0150  10.1391
    1.0334  10.3167
    >>> _find_positions_in_between(
    ...     source_series=end_mean_curve, search_value=target_end_y_value,
    ... )
    (4, 5)
    >>> index_for_target_y = _calculate_cutting_index_for_value(
    ...     series_to_cut=end_mean_curve,
    ...     cutting_y_value=target_end_y_value,
    ... )
    >>> round(index_for_target_y, 6)
    1.015833
    """
    sample_curves = copy.copy(_current_test_curves)
    extrapolates = FrozenStdExtrapolation(
        use_previous_iteration=False, target_threshold=0.0001
    )
    a_mean_curve = ArithmeticMeanCurve(sample_curves, method=extrapolates)
    target_end_x_value = a_mean_curve.statistics.end_x_mean
    target_end_y_value = a_mean_curve.statistics.end_y_mean

    a_mean_curve.prepare_extrapolation()
    prepared_end_cap = _prepare_end_cap_for_extrapolation(a_mean_curve=a_mean_curve)

    last_full_value_row = a_mean_curve.get_last_full_row()
    last_value_row_mean_value = last_full_value_row.mean(axis=1)
    starting_x_mean_value = last_value_row_mean_value.iloc[0]

    extrapolated_end_cap = _extrapolate_until_targets_are_included(
        extrapolates=a_mean_curve,
        end_cap=prepared_end_cap,
        target_x_mean_value=target_end_x_value,
        target_y_mean_value=target_end_y_value,
        starting_last_mean_value=starting_x_mean_value,
    )
    extended_extrapolated_end_cap = pandas.concat(
        [last_full_value_row, extrapolated_end_cap]
    )
    current_end_mean_curve = extended_extrapolated_end_cap.mean(axis=1)
    return current_end_mean_curve, target_end_y_value


def test_calculate_cutting_index_for_value():
    """
    _calculate_cutting_index_for_value leads to a wrong cutting index,
    if FrozenStdExtrapolation is run with use_previous_iteration=False.

    >>> from arithmeticmeancurve import _find_positions_in_between
    >>> import examplecurves
    >>> from doctestprinter import doctest_print
    >>> indexes, values, target_end_y_value = test_calculate_cutting_index_for_value()
    >>> target_end_y_value
    10.147148237071187
    >>> indexes
    [0.99814, 1.01435, 1.01452, 1.01454, 1.015, 1.03342]
    >>> values
    [9.97995, 10.1423, 10.14135, 10.14009, 10.13908, 10.31668]
    >>> from pandas import Series
    >>> sample_series = Series(values, index=indexes)
    >>> sample_search_value = target_end_y_value
    >>> _find_positions_in_between(sample_series, sample_search_value)
    (4, 5)
    """
    end_mean_curve, target_end_y_value = test_find_cutting_index_2()
    indexes = [round(x_i, 5) for x_i in end_mean_curve.index]
    return indexes, end_mean_curve.round(5).to_list(), target_end_y_value


def test_do_not_use_previous_iteration_mean_curve():
    """
    >>> import examplecurves
    >>> from doctestprinter import print_pandas
    >>> mean_curve, std_curve = test_do_not_use_previous_iteration_mean_curve()
    >>> print_pandas(mean_curve, formats="{:>.4f}" )
    x
    0.0000   0.0000
    0.2495   2.4950
    0.2536   2.5355
    0.2536   2.5359
    0.2536   2.5360
    0.2584   2.5832
    0.4991   4.9900
    0.5072   5.0710
    0.5073   5.0718
    0.5073   5.0720
    0.5167   5.1663
    0.7486   7.4850
    0.7608   7.6065
    0.7609   7.6078
    0.7609   7.6079
    0.7751   7.7495
    0.9981   9.9800
    1.0145  10.1414
    1.0145  10.1401
    1.0150  10.1391
    1.0158  10.1471
    """
    sample_curves = copy.copy(_current_test_curves)
    extrapolates = FrozenStdExtrapolation(
        use_previous_iteration=False, target_threshold=0.0001
    )
    a_mean_curve = ArithmeticMeanCurve(sample_curves, method=extrapolates)
    return _calculate_mean_and_std_curve(a_mean_curve)